# Test added by zzt (Defender)
from random import randint

from helper import *


@TestInstance
def test_scdata():
    # FIXME
    # with expect_eperror():
    #     TrgUnit(233)
    TrgUnit("Terran Vulture").groundWeapon.damage += 3
    marine = TrgUnit("Terran Marine")
    marine = TrgUnit(marine)  # copy inner value
    marine = TrgUnit.cast(marine)  # share reference
    test_equality("TrgUnit(marine) = 0", marine, 0)
    test_equality("marine.maxHp = 40 * 256", marine.maxHp, 40 * 256)
    test_equality(
        "marine.maxHp = EUDVar(40 * 256)", marine.maxHp, EUDVariable(40 * 256)
    )
    test_assert("marine.maxHp = 40 * 256 conditional", marine.maxHp == 40 * 256)

    one = EUDVariable(1)
    trigcount = GetTriggerCounter()
    ghost = TrgUnit(one)  # copy variable one
    ep_assert(
        trigcount + 1 == GetTriggerCounter(),
        f"Trigger count mismatch: {trigcount} + 1 != {GetTriggerCounter()}",
    )
    test_equality(
        "TrgUnit(EUDVar(ghost)).maxHp = EUDVar(45 * 256)",
        ghost.maxHp,
        EUDVariable(45 * 256),
    )
    test_assert(
        "TrgUnit(EUDVar(ghost)).maxHp = EUDVar(45 * 256) conditional",
        ghost.maxHp == 45 * 256,
    )

    trigcount = GetTriggerCounter()
    ghost_cast = TrgUnit.cast(one)  # reuse variable
    ep_assert(
        trigcount == GetTriggerCounter(),
        f"Trigger count mismatch: {trigcount} != {GetTriggerCounter()}",
    )
    ep_assert(one is not ghost._value, "New instance should not alias")
    ep_assert(one is ghost_cast._value, "cast should alias")
    one << 2
    test_equality(
        "TrgUnit(eudvar).maxHp, check robustness to variable change",
        ghost.maxHp,
        EUDVariable(45 * 256),
    )
    test_equality(
        "TrgUnit(eudvar).maxHp, referencing variable",
        ghost_cast.maxHp,
        EUDVariable(80 * 256),
    )
    test_assert(
        "TrgUnit(eudvar).maxHp conditional",
        ghost_cast.maxHp == 80 * 256,
    )

    test_equality("arithmetic on UnitData", ghost + 3, 4)
    ghost << 2
    ep_assert(isinstance(ghost, TrgUnit), "preserve type on assignment")
    test_equality(
        "TrgUnit(eudvar).maxHp, after assignment",
        ghost_cast.maxHp,
        EUDVariable(80 * 256),
    )
    ghost -= 1
    ep_assert(isinstance(ghost, TrgUnit), "preserve type after in-place operations")

    zealot_data = TrgUnit("Protoss Zealot")

    previous_value = zealot_data.maxHp
    zealot_data.maxHp = 80 * 256

    test_equality(
        "TrgUnit.maxHp, check if read/write points to the same address",
        zealot_data.maxHp,
        80 * 256,
    )

    test_equality(
        "TrgUnit.maxHp, check if write writes to the correct address",
        zealot_data.maxHp,
        f_dwread(0x662350 + 4 * TrgUnit("Protoss Zealot")),
    )

    zealot_data.maxHp = previous_value

    previous_value = TrgUnit("Goliath Turret").maxHp
    TrgUnit("Goliath Turret").maxHp = 512

    test_equality(
        "TrgUnit.subUnit, check UnitData to int conversion",
        TrgUnit("Terran Goliath").subUnit,
        TrgUnit("Goliath Turret"),
    )
    test_assert(
        "TrgUnit.subUnit conditional",
        TrgUnit("Terran Goliath").subUnit == TrgUnit("Goliath Turret"),
    )
    test_equality(
        "TrgUnit.subUnit, check if member type of unit works",
        TrgUnit("Terran Goliath").subUnit.maxHp,
        512,
    )
    test_assert(
        "TrgUnit.subUnit.maxHp conditional",
        TrgUnit("Terran Goliath").subUnit.maxHp == 512,
    )

    test_equality(
        "TrgUnit.subUnit, check if chain works",
        TrgUnit("Terran Goliath").subUnit.subUnit,
        228,
    )
    test_assert(
        "TrgUnit.subUnit.subUnit conditional",
        TrgUnit("Terran Goliath").subUnit.subUnit == 228,
    )

    test_equality(
        "TrgUnit.flingy, check if chain through other data types works",
        TrgUnit("Protoss Dragoon").flingy,
        Flingy("Dragoon"),
    )
    test_assert(
        "TrgUnit.flingy conditional",
        TrgUnit("Protoss Dragoon").flingy == Flingy("Dragoon"),
    )

    test_equality(
        "TrgUnit.flingy.sprite, check if chain through other data types works",
        TrgUnit("Zerg Zergling").flingy.sprite,
        Sprite("Zergling"),
    )
    test_assert(
        "TrgUnit.flingy.sprite conditional",
        TrgUnit("Zerg Zergling").flingy.sprite == Sprite("Zergling"),
    )

    test_equality(
        "TrgUnit.flingy.sprite.image, check if chain through other data types works",
        TrgUnit("Terran Marine").flingy.sprite.image,
        Image("Marine"),
    )
    test_assert(
        "TrgUnit.flingy.sprite.image conditional",
        TrgUnit("Terran Marine").flingy.sprite.image == Image("Marine"),
    )

    archon_variable = EUDVariable()
    archon_variable << TrgUnit("Protoss Archon")
    archon = TrgUnit(archon_variable)

    test_equality(
        "TrgUnit.flingy, check if chain from variable works",
        archon.flingy,
        Flingy("Archon Energy"),
    )
    test_assert(
        "TrgUnit.flingy conditional", archon.flingy == Flingy("Archon Energy")
    )

    test_equality(
        "TrgUnit.flingy.sprite, check if chain from variable works",
        archon.flingy.sprite,
        Sprite("Archon Energy"),
    )
    test_assert(
        "TrgUnit.flingy.sprite conditional",
        archon.flingy.sprite == Sprite("Archon Energy"),
    )

    test_equality(
        "TrgUnit.flingy.sprite.image, check if chain from variable works",
        archon.flingy.sprite.image,
        Image("Archon Energy"),
    )
    test_assert(
        "TrgUnit.flingy.sprite.image conditional",
        archon.flingy.sprite.image == Image("Archon Energy"),
    )

    TrgUnit("Goliath Turret").maxHp = previous_value
    DoActions(SetResources(P7, SetTo, 100, OreAndGas))

    test_equality(
        "TrgPlayer.ore, check ore, gas amounts read", [P7.ore, P7.gas], [100, 100]
    )
    test_assert(
        "TrgPlayer.ore/gas conditional",
        [P7.ore == 100, P7.gas == 100],
    )

    P7.ore += 200

    test_equality("TrgPlayer.ore, check ore amount write", P7.ore, 300)
    test_assert("TrgPlayer.ore, check Accumulate", Accumulate(P7, Exactly, 300, Ore))
    DoActions(SetResources(P7, SetTo, 0, OreAndGas))

    with expect_eperror():
        TrgUnit("zzt, GOD OF EUD")

    scv = TrgUnit("Terran SCV")
    test_equality("scv.baseProperty flags", scv.baseProperty, 0x58010008)
    test_assert("scv.baseProperty flags conditional", scv.baseProperty == 0x58010008)
    scv.baseProperty.Hero = True
    test_equality(
        "scv.baseProperty individual flags",
        [
            scv.baseProperty.Worker,
            scv.baseProperty.AutoAttackAndMove,
            scv.baseProperty.CanAttack,
            scv.baseProperty.Mechanical,
            scv.baseProperty.Hero,
            scv.baseProperty.ResourceDepot,
        ],
        [
            True,
            True,
            True,
            True,
            True,
            False,
        ],
    )
    test_assert(
        "scv.baseProperty individual flags conditional",
        [
            scv.baseProperty.Worker == True,
            scv.baseProperty.AutoAttackAndMove == True,
            scv.baseProperty.CanAttack == True,
            scv.baseProperty.Mechanical == True,
            scv.baseProperty.Hero == True,
            scv.baseProperty.ResourceDepot == False,
        ],
    )

    test_equality("scv.nameString == 0", scv.nameString, 0)
    name = EncodeString("dpdkfah!")  # EncodeString does string interning
    scv.nameString = "dpdkfah!"
    test_equality("scv.nameString test", scv.nameString, name)
    # FIXME
    # with expect_eperror():
    #     scv.nameString = 65536

    test_equality(
        "scv.dontBecomeGuard == True",
        scv.dontBecomeGuard,
        True,
    )
    scv.dontBecomeGuard = False
    test_equality(
        "scv.dontBecomeGuard == False",
        scv.dontBecomeGuard,
        False,
    )
    with expect_error(NotImplementedError):
        scv.iaddattr("dontBecomeGuard", 1)

    P1.unitColor = 123
    test_equality("P1.unitColor = 123", P1.unitColor, 123)
    test_assert("P1.unitColor = 123 conditional", P1.unitColor == 123)
    test_equality("P1.unitColor", P1.unitColor, f_bread(0x581D76))
    P2.unitColor = 234
    test_equality("P2.unitColor = 234", P2.unitColor, 234)

    P1.unitColor = 111
    P2.unitColor = 165


@TestInstance
def test_scdata_reference():
    goliath_cunit = CUnit.from_next()
    DoActions(CreateUnit(1, "Terran Goliath", "Anywhere", P8))
    test_equality(
        "Cunit subunit type check",
        goliath_cunit.subUnit.unitType,
        TrgUnit("Goliath Turret"),
    )

    DoActions(RemoveUnit("Terran Goliath", P8))


@TestInstance
def test_scdata_caching():
    from eudplib.memio import muldiv4table
    from eudplib.scdata.offsetmap import EPDOffsetMap

    ptr = EUDVariable()

    @EUDFunc
    def foo():
        return CUnit.from_ptr(ptr)

    test_equality("caching skip test", foo(), 0)

    epds = []
    ptr << 0x58A364 + 4
    epds.append(foo())

    ptr << 0x58A364
    epds.append(foo())

    ptr << 0x58A364 + 8
    epds.append(foo())

    ptr << 0
    epds.append(foo())
    test_equality("caching tests", epds, [1, 0, 2, 0])

    p = TrgPlayer.cast(EUDVariable())
    v = p._value
    p << P8
    P8.minimapColor = 128
    color = p.minimapColor
    update_start, update_restore, update_end = EPDOffsetMap._update[v]
    test_equality("P8.minimapColor = 128", color, 128)
    test_equality("Cache invalidation", f_dwread(update_start + 4), update_restore)

    P8.unitColor = 178
    test_equality("P8.unitColor = 178", p.unitColor, 178)
    test_equality(
        "Reuse cached values",
        f_dwread(update_start + 4),
        f_dwread(update_start + 348),
    )

    u = TrgUnit.cast(v)
    u << "Alan Schezar"
    test_equality("Alan Schezar.timeCost", u.timeCost, 1200)
    test_equality("Cache invalidation", f_dwread(update_start + 4), update_restore)
    test_equality(
        "varcount", EUDVariable(update_end), muldiv4table.muldiv_end_table[5]
    )

    # += 1 is specialized to update cached values
    u += 1
    f_simpleprint("VAL:", u, ", CACHE_COND:", f_dwread(update_start + 16))
    test_equality("Alan Schezar += 1; .flingy", u.flingy, Flingy("Goliath Turret"))
    test_equality(
        "Reuse cached values",
        f_dwread(update_start + 4),
        f_dwread(update_start + 348),
    )

    P8.unitColor = 135
    P8.minimapColor = 135


@TestInstance
def test_scdata_value_range_extension():
    v = EUDVariable(0x17)  # 23

    p = TrgPlayer.cast(v)
    P12.cumulativeMineral = 9999  # 0x57F1AC = 0x57F150 + 0x17 * 4
    test_equality("TrgPlayer with bitmask 0xFF", p.cumulativeGas, 9999)

    u = TrgUnit.cast(v)
    # TrgUnit usage will extend bitmask of `v` from 0xF to 0xFF
    # and this applies retroactively to every usages of `v`
    test_equality("TrgUnit with bitmask 0xFF", u.gasCost, 200)

    P12.cumulativeMineral = 0


@TestInstance
def test_scdata_var():
    p = TrgPlayer.cast(EUDVariable(4))
    test_equality("P5 Color = 156", [p.minimapColor, p.unitColor], [156, 156])
    p += 1
    test_equality("P6 Color = 19", [p.minimapColor, p.unitColor], [19, 19])
    p += 1
    test_equality("P7 Color = 84", [p.minimapColor, p.unitColor], [84, 84])
    p += 1
    test_equality("P8 Color = 135", [p.minimapColor, p.unitColor], [135, 135])
    p += 1
    test_equality("P9 Color = 185", [p.minimapColor, p.unitColor], [185, 185])


@TestInstance
def test_scdata_current_upgrade():
    for upgrade_id in (Upgrade("Protoss Plasma Shields"), Upgrade("Charon Booster")):
        upgrades = [Upgrade(upgrade_id), Upgrade(EUDVariable(upgrade_id))]
        players = [P11, TrgPlayer(EUDVariable(10))]
        if upgrade_id < 46:
            ptr = 0x58D2B0 + upgrade_id + 46 * 10
        else:
            ptr = 0x58F32C + upgrade_id - 46 + 15 * 10
        for upgrade, player in zip(upgrades, players):
            lv = randint(1, 255)
            u = "var" if IsEUDVariable(upgrade) else "const"
            p = "var" if IsEUDVariable(player) else "const"
            f_bwrite(ptr, lv)
            test_equality(f"lv = Upgrade({u})[TrgPlayer({p})]", upgrade[player], lv)

            lv = randint(1, 255)
            upgrade[player] = lv
            test_equality(f"Upgrade({u})[TrgPlayer({p})] = lv", f_bread(ptr), lv)

        f_bwrite(ptr, 0)

    for tech_id in (Tech("Burrowing"), Tech("Feedback")):
        techs = [Tech(tech_id), Tech(EUDVariable(tech_id))]
        players = [P12, TrgPlayer(EUDVariable(11))]
        if tech_id < 24:
            ptr = 0x58CF44 + tech_id + 24 * 11
        else:
            ptr = 0x58F140 + tech_id - 24 + 20 * 11
        for tech, player in zip(techs, players):
            lv = 1
            u = "var" if IsEUDVariable(tech) else "const"
            p = "var" if IsEUDVariable(player) else "const"
            f_bwrite(ptr, lv)
            test_equality(f"lv = Tech({u})[TrgPlayer({p})]", tech[player], lv)

            lv = 0
            tech[player] = lv
            test_equality(f"Tech({u})[TrgPlayer({p})] = lv", f_bread(ptr), lv)



# Copyright 2022 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ... import core as c
from ... import ctrlstru as cs


@c.EUDFunc
def _pow(a, b):
    ret, _2n = c.EUDCreateVariables(2)
    c.SetVariables([ret, _2n], [1, 1])
    # 2^n < b 인 모든 a^(2^n) 구하기
    if cs.EUDWhile()(_2n <= b):
        # b에 (2^n)이 있으면 답에 a^(2^n)을 곱한다
        if cs.EUDIf()(b.AtLeastX(1, _2n)):
            ret *= a
        cs.EUDEndIf()
        _2n += _2n
        a *= a
    cs.EUDEndWhile()
    return ret


def f_pow(a, b):
    """
    f_pow(a, b) calculates a ** b
    """
    if isinstance(a, int) and isinstance(b, int):
        return a**b
    return _pow(a, b)



from helper import *


@TestInstance
def test_listloop_compiles():
    if EUDIf()(Never()):
        a, b = EUDCreateVariables(2)
        for ptr, epd in EUDLoopList(a, b):
            pass
    EUDEndIf()



from __future__ import annotations

import datetime

import pytest

import env  # noqa: F401
from pybind11_tests import chrono as m


def test_chrono_system_clock():
    # Get the time from both c++ and datetime
    date0 = datetime.datetime.today()
    date1 = m.test_chrono1()
    date2 = datetime.datetime.today()

    # The returned value should be a datetime
    assert isinstance(date1, datetime.datetime)

    # The numbers should vary by a very small amount (time it took to execute)
    diff_python = abs(date2 - date0)
    diff = abs(date1 - date2)

    # There should never be a days difference
    assert diff.days == 0

    # Since datetime.datetime.today() calls time.time(), and on some platforms
    # that has 1 second accuracy, we compare this way
    assert diff.seconds <= diff_python.seconds


def test_chrono_system_clock_roundtrip():
    date1 = datetime.datetime.today()

    # Roundtrip the time
    date2 = m.test_chrono2(date1)

    # The returned value should be a datetime
    assert isinstance(date2, datetime.datetime)

    # They should be identical (no information lost on roundtrip)
    diff = abs(date1 - date2)
    assert diff == datetime.timedelta(0)


def test_chrono_system_clock_roundtrip_date():
    date1 = datetime.date.today()

    # Roundtrip the time
    datetime2 = m.test_chrono2(date1)
    date2 = datetime2.date()
    time2 = datetime2.time()

    # The returned value should be a datetime
    assert isinstance(datetime2, datetime.datetime)
    assert isinstance(date2, datetime.date)
    assert isinstance(time2, datetime.time)

    # They should be identical (no information lost on roundtrip)
    diff = abs(date1 - date2)
    assert diff.days == 0
    assert diff.seconds == 0
    assert diff.microseconds == 0

    # Year, Month & Day should be the same after the round trip
    assert date1 == date2

    # There should be no time information
    assert time2.hour == 0
    assert time2.minute == 0
    assert time2.second == 0
    assert time2.microsecond == 0


SKIP_TZ_ENV_ON_WIN = pytest.mark.skipif(
    "env.WIN", reason="TZ environment variable only supported on POSIX"
)


@pytest.mark.parametrize(
    "time1",
    [
        datetime.datetime.today().time(),
        datetime.time(0, 0, 0),
        datetime.time(0, 0, 0, 1),
        datetime.time(0, 28, 45, 109827),
        datetime.time(0, 59, 59, 999999),
        datetime.time(1, 0, 0),
        datetime.time(5, 59, 59, 0),
        datetime.time(5, 59, 59, 1),
    ],
)
@pytest.mark.parametrize(
    "tz",
    [
        None,
        pytest.param("Europe/Brussels", marks=SKIP_TZ_ENV_ON_WIN),
        pytest.param("Asia/Pyongyang", marks=SKIP_TZ_ENV_ON_WIN),
        pytest.param("America/New_York", marks=SKIP_TZ_ENV_ON_WIN),
    ],
)
def test_chrono_system_clock_roundtrip_time(time1, tz, monkeypatch):
    if tz is not None:
        monkeypatch.setenv("TZ", f"/usr/share/zoneinfo/{tz}")

    # Roundtrip the time
    datetime2 = m.test_chrono2(time1)
    date2 = datetime2.date()
    time2 = datetime2.time()

    # The returned value should be a datetime
    assert isinstance(datetime2, datetime.datetime)
    assert isinstance(date2, datetime.date)
    assert isinstance(time2, datetime.time)

    # Hour, Minute, Second & Microsecond should be the same after the round trip
    assert time1 == time2

    # There should be no date information (i.e. date = python base date)
    assert date2.year == 1970
    assert date2.month == 1
    assert date2.day == 1


def test_chrono_duration_roundtrip():
    # Get the difference between two times (a timedelta)
    date1 = datetime.datetime.today()
    date2 = datetime.datetime.today()
    diff = date2 - date1

    # Make sure this is a timedelta
    assert isinstance(diff, datetime.timedelta)

    cpp_diff = m.test_chrono3(diff)

    assert cpp_diff == diff

    # Negative timedelta roundtrip
    diff = datetime.timedelta(microseconds=-1)
    cpp_diff = m.test_chrono3(diff)

    assert cpp_diff == diff


def test_chrono_duration_subtraction_equivalence():
    date1 = datetime.datetime.today()
    date2 = datetime.datetime.today()

    diff = date2 - date1
    cpp_diff = m.test_chrono4(date2, date1)

    assert cpp_diff == diff


def test_chrono_duration_subtraction_equivalence_date():
    date1 = datetime.date.today()
    date2 = datetime.date.today()

    diff = date2 - date1
    cpp_diff = m.test_chrono4(date2, date1)

    assert cpp_diff == diff


def test_chrono_steady_clock():
    time1 = m.test_chrono5()
    assert isinstance(time1, datetime.timedelta)


def test_chrono_steady_clock_roundtrip():
    time1 = datetime.timedelta(days=10, seconds=10, microseconds=100)
    time2 = m.test_chrono6(time1)

    assert isinstance(time2, datetime.timedelta)

    # They should be identical (no information lost on roundtrip)
    assert time1 == time2


def test_floating_point_duration():
    # Test using a floating point number in seconds
    time = m.test_chrono7(35.525123)

    assert isinstance(time, datetime.timedelta)

    assert time.seconds == 35
    assert 525122 <= time.microseconds <= 525123

    diff = m.test_chrono_float_diff(43.789012, 1.123456)
    assert diff.seconds == 42
    assert 665556 <= diff.microseconds <= 665557


def test_nano_timepoint():
    time = datetime.datetime.now()
    time1 = m.test_nano_timepoint(time, datetime.timedelta(seconds=60))
    assert time1 == time + datetime.timedelta(seconds=60)


def test_chrono_different_resolutions():
    resolutions = m.different_resolutions()
    time = datetime.datetime.now()
    resolutions.timestamp_h = time
    resolutions.timestamp_m = time
    resolutions.timestamp_s = time
    resolutions.timestamp_ms = time
    resolutions.timestamp_us = time



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from typing import TypeAlias

from .. import core as c
from .. import utils as ut
from ..core import ConstExpr, EUDVariable, RlocInt_C
from ..core.eudfunc.eudf import _EUDPredefineParam
from ..core.rawtrigger.constenc import SetTo, TrgModifier

_lowordfilter = c.EUDXVariable(0, SetTo, 0, 0xFFFF)
_lsbytefilter = c.EUDXVariable(0, SetTo, 0, 0xFF)
_lobytefilter = c.EUDXVariable(0, SetTo, 0, 0xFF00)
_hibytefilter = c.EUDXVariable(0, SetTo, 0, 0xFF0000)
_msbytefilter = c.EUDXVariable(0, SetTo, 0, 0xFF000000)

Constant: TypeAlias = ConstExpr | int | RlocInt_C
_fillers_use_seqcompute: list[
    tuple[
        Constant | EUDVariable,
        TrgModifier | None,
        Constant | EUDVariable,
    ]
] = []


def _flush_filler():
    global _fillers_use_seqcompute
    if _fillers_use_seqcompute:
        c.SeqCompute(_fillers_use_seqcompute)
        _fillers_use_seqcompute.clear()


def _filldw(dstepd: Constant | EUDVariable, v1: Constant | EUDVariable) -> None:
    _fillers_use_seqcompute.append((dstepd, c.SetTo, v1))


def _fillloword(dstepd: Constant, v1: EUDVariable) -> None:
    _fillers_use_seqcompute.extend(
        (
            (_lowordfilter, c.SetTo, v1),
            (dstepd, None, _lowordfilter),
        )
    )


def _filllsbyte(dstepd: Constant, v1: EUDVariable) -> None:
    _fillers_use_seqcompute.extend(
        (
            (_lsbytefilter, c.SetTo, v1),
            (dstepd, None, _lsbytefilter),
        )
    )


@_EUDPredefineParam(1)
@c.EUDFunc
def _fill_b__(v1) -> None:
    _lobytefilter << 0
    for i in ut._rand_lst(range(8)):
        c.RawTrigger(
            conditions=v1.AtLeastX(1, 2**i),
            actions=_lobytefilter.AddNumber(2 ** (i + 8)),
        )


def _filllobyte(dstepd: Constant, v1: Constant | EUDVariable) -> None:
    _flush_filler()
    _fill_b__(v1)
    _fillers_use_seqcompute.append((dstepd, None, _lobytefilter))


@_EUDPredefineParam(1)
@c.EUDFunc
def _fill__b_(v1) -> None:
    _hibytefilter << 0
    for i in ut._rand_lst(range(8)):
        c.RawTrigger(
            conditions=v1.AtLeastX(1, 2**i),
            actions=_hibytefilter.AddNumber(2 ** (i + 16)),
        )


def _fillhibyte(dstepd: Constant, v1: Constant | EUDVariable) -> None:
    _flush_filler()
    _fill__b_(v1)
    _fillers_use_seqcompute.append((dstepd, None, _hibytefilter))


@_EUDPredefineParam(1)
@c.EUDFunc
def _fill___b(v1) -> None:
    _msbytefilter << 0
    for i in ut._rand_lst(range(8)):
        c.RawTrigger(
            conditions=v1.AtLeastX(1, 2**i),
            actions=_msbytefilter.AddNumber(2 ** (i + 24)),
        )


def _fillmsbyte(dstepd: Constant, v1: Constant | EUDVariable) -> None:
    _flush_filler()
    _fill___b(v1)
    _fillers_use_seqcompute.append((dstepd, None, _msbytefilter))



import os
import sys

sys.path.insert(1, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import re
import types

import eudplib

module_to_doc = eudplib

exclude_types = [dict, str, types.ModuleType]
eudftypes = [eudplib.EUDFuncN]
allowed_type = eudftypes + [type, types.FunctionType]
exclude_names = ["__loader__", "__path__", "__spec__"]
section_header_charr = ["=", "-", "'", "~", "^"]


# Get list of already documented items
print("Getting list of currently documented structures")
with open("api.rst", "r", encoding="utf-8") as doc_file:
    rstinputs = doc_file.read().split("\n")
documented_functions = set()

func_regex = re.compile(r"-   \*\*(.+)?\*\*")

for line in rstinputs:
    func_match_result = func_regex.match(line)
    if func_match_result:
        funcname = func_match_result.groups(1)[0]
        print(funcname)
        if funcname in documented_functions:
            print("[Warning] Document duplication of function %s." % funcname)
        documented_functions.add(funcname)
        continue

print("Collecting list of documentation-needed structures")
doc_needed_functions = set()
fc_documented = {}


# Document module
for name, value in module_to_doc.__dict__.items():
    # Exclude list
    if name in exclude_names or type(value) in exclude_types:
        continue

    if not any(isinstance(value, t) for t in allowed_type):
        continue

    if value.__doc__ is None:
        print(" [Warning] undocumented value %s" % name)
        documented = False

    else:
        documented = True

    if isinstance(value, types.FunctionType) or isinstance(value, eudplib.EUDFuncN):
        doc_needed_functions.add(name)
        fc_documented[name] = documented

    elif type(value) is type:
        doc_needed_functions.add(name)
        fc_documented[name] = documented

print("\n==================================\n")


# Summary
"""
print('Current entries:')
print('  Functions:')
for k in doc_needed_functions:
    print('    - %s' % k)
print('  Classes:')
for k in doc_needed_classes:
    print('    - %s' % k)
print('==================================')
"""

# Diff
unused_functions = documented_functions.difference(doc_needed_functions)

if unused_functions:
    print("Entrys now unused:")
    if unused_functions:
        print("  Unused functions:")
        print(unused_functions)
        for k in sorted(list(unused_functions)):
            print("    - %s" % k)

    print("\n==================================\n")


# Diff2
used_functions = doc_needed_functions.difference(documented_functions)

if used_functions:
    print("New entries:")
    for k in sorted(list(used_functions)):
        print("    - %s %s" % (k, "(Undocumented)" if not fc_documented[k] else ""))

    print("\n==================================\n")

print("Diff done")



from __future__ import annotations

import os

import pytest

import env  # noqa: F401
from pybind11_tests import eval_ as m


def test_evals(capture):
    with capture:
        assert m.test_eval_statements()
    assert capture == "Hello World!"

    assert m.test_eval()
    assert m.test_eval_single_statement()

    assert m.test_eval_failure()


@pytest.mark.xfail("env.PYPY", raises=RuntimeError)
def test_eval_file():
    filename = os.path.join(os.path.dirname(__file__), "test_eval_call.py")
    assert m.test_eval_file(filename)

    assert m.test_eval_file_failure()


def test_eval_empty_globals():
    assert "__builtins__" in m.eval_empty_globals(None)

    g = {}
    assert "__builtins__" in m.eval_empty_globals(g)
    assert "__builtins__" in g


def test_eval_closure():
    global_, local = m.test_eval_closure()

    assert global_["closure_value"] == 42
    assert local["closure_value"] == 0

    assert "local_value" not in global_
    assert local["local_value"] == 0

    assert "func_global" not in global_
    assert local["func_global"]() == 42

    assert "func_local" not in global_
    with pytest.raises(NameError):
        local["func_local"]()



from ..core import (
    Db,
    EUDFunc,
    EUDReturn,
    EUDVariable,
    Exactly,
    Memory,
    RawTrigger,
    SetMemory,
    SetTo,
)
from ..ctrlstru import (
    EUDEndSwitch,
    EUDSwitch,
    EUDSwitchCase,
    EUDSwitchDefault,
)
from ..scdata import AllPlayers
from ..utils import EPD, ExprProxy, ep_assert
from .cpprint import f_raise_CCMU
from .strcommon import epd2s

LOCALES = (
    "enUS",  # [1] English
    "frFR",  # [2] Français
    "itIT",  # [3] Italiano
    "deDE",  # [4] Deutsch
    "esES",  # [5] Español - España
    "esMX",  # [6] Español - Latino
    "ptBR",  # [7] Português
    "zhCN",  # [8] 简体中文
    "zhTW",  # [9] 繁體中文
    "jaJP",  # [10] 日本語
    "koUS",  # [11] 한국어 (음역)
    "koKR",  # [12] 한국어 (완역)
    "plPL",  # [13] Polski
    "ruRU",  # [14] Русский
)


class _LocalLocale(ExprProxy):
    dont_flatten = True

    def __init__(self, initvar) -> None:
        super().__init__(initvar)

    def __lshift__(self, other):
        if isinstance(other, str):
            ep_assert(other in LOCALES)
            other = 1 + LOCALES.index(other)
        return super().__lshift__(other)

    def __eq__(self, other):
        if isinstance(other, str):
            ep_assert(
                other in LOCALES,
                f"Invalid locale: {other}\nAvailable locales: {LOCALES}",
            )
            other = 1 + LOCALES.index(other)
        return super().__eq__(other)

    def __ne__(self, other):
        if isinstance(other, str):
            ep_assert(other in LOCALES)
            other = 1 + LOCALES.index(other)
        return super().__ne__(other)

    @staticmethod
    @EUDFunc
    def _fmt(locale):
        EUDSwitch(locale)
        for i, _locale in enumerate(LOCALES):
            if EUDSwitchCase()(i + 1):
                EUDReturn(EPD(Db(_locale)))
        if EUDSwitchDefault()():
            EUDReturn(EPD(Db("(unknown locale)")))
        EUDEndSwitch()

    def fmt(self):
        return epd2s(_LocalLocale._fmt(self))


LocalLocale = _LocalLocale(EUDVariable(0))
_MAGIC_NUMBERS = {
    "enUS": (1650552405, 1948280172, 1919098991),
    "frFR": (1869638985, 1651078003, 1679844716),
    "itIT": (1869638985, 1651078003, 543517801),
    "deDE": (543516996, 1752066373, 544500069),
    "esES": (1931505486, 1634213989, 1685024800),
    "esMX": (1931505486, 1970282597, 543515749),
    "ptBR": (1869638985, 2915267443, 543974774),
    "zhCN": (3869284326, 2296747443, 3132876187),
    "zhTW": (3869344999, 3152385459, 2692803002),
    "jaJP": (3819340771, 2212727683, 2290344835),
    "koUS": (3953171692, 2649529227, 2213290116),
    "koKR": (3953171692, 2649529227, 2213290116),
    "plPL": (543517006, 3167055725, 1931501934),
    "ruRU": (3050347984, 3498299680, 3501248692),
}


def _detect_locale():
    f_raise_CCMU(AllPlayers)
    status_msg = 0x640B60 + 218 * 12
    for locale_id, locale in enumerate(LOCALES):
        magic_numbers = _MAGIC_NUMBERS[locale]
        # FIXME: Can't disambiguate between koUS and koKR
        if locale == "koKR":
            continue
        RawTrigger(
            conditions=[
                Memory(status_msg, Exactly, magic_numbers[0]),
                Memory(status_msg + 4, Exactly, magic_numbers[1]),
                Memory(status_msg + 8, Exactly, magic_numbers[2]),
            ],
            actions=LocalLocale.SetNumber(locale_id + 1),
        )
    # clear error message
    RawTrigger(actions=SetMemory(status_msg, SetTo, 0))



import inspect
import os
import sys

sys.path.insert(0, os.path.abspath("..\\"))

import eudplib

eudftype = eudplib.core.varfunc.eudf.EUDFuncN


def varnames(f):
    if isinstance(f, eudftype):
        f = f._fdecl_func
    argcount = f.__code__.co_argcount
    return f.__code__.co_varnames[:argcount]


def isFunction(f):
    return isinstance(f, eudftype) or inspect.isfunction(f)


deflist = []
for name in eudplib.__all__:
    value = eudplib.__dict__[name]
    if isFunction(value):
        deflist.append("%s(%s)" % (name, ", ".join(varnames(value))))

deflist.sort()

for d in deflist:
    print(d)



from helper import *


@EUDFunc
def f_add(a, b):
    return a + b


@EUDFunc
def f_mul(a, b):
    return a * b


@EUDFunc
def f_div(a, b):
    return a // b, a % b


@EUDFunc
def f_addmul(a, b):
    return a + b, a * b


noretCheck = EUDVariable()


@EUDFunc
def testnoret():
    noretCheck << 1


@TestInstance
def test_fptr():
    p = EUDFuncPtr(2, 1)()

    # Test - Simple pattern
    p << f_add
    a = p(7, 9)

    p << f_mul
    b = p(7, 9)

    with expect_eperror():
        p << f_div  # Should trigger error

    test_equality("Simple fptr call", [a, b], [16, 63])

    # Test2 - multi return
    q = EUDFuncPtr(2, 2)(f_div)
    a1, a2 = q(12, 5)
    q << f_addmul
    b1, b2 = q(12, 5)
    q << f_div
    c1, c2 = q(17, 3)
    test_equality(
        "Function pointer test", [a1, a2, b1, b2, c1, c2], [2, 2, 17, 60, 5, 2]
    )

    # Test 3 - no arg no ret
    r = EUDFuncPtr(0, 0)(testnoret)
    r()
    test_assert("Function pointer test (0.0 Func)", noretCheck == 1)

    # Test 4 - transfer fptr to fptr
    r = EUDFuncPtr(2, 1).cast(p)
    p << f_mul
    a, b = p(7, 9), r(7, 9)
    r << f_add
    c, d = p(7, 9), r(7, 9)
    test_equality("Indirect fptr call", [a, b, c, d], [63, 63, 16, 16])

    # Test 5 - fptr cloning
    s = p.copy()
    p << f_mul
    e, f = p(7, 9), s(7, 9)
    test_equality("Fptr cloning", [e, f], [63, 16])


# Nested fptr call
@EUDTypedFunc([EUDFuncPtr(1, 1)])
def f1(f):
    return 4 + f(4) + 6


@EUDFunc
def f(x):
    return x * x


@TestInstance
def test_nested_fptr():
    k = EUDFuncPtr(1, 1)(f1)
    test_equality("Nested fptr", k(EUDFuncPtr(1, 1)(f)), 26)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Iterator

from ... import core as c
from ... import ctrlstru as cs
from ... import utils as ut
from ...localize import _
from ...memio import (
    f_bread_cp,
    f_cunitepdread_epd,
    f_dwepdread_epd,
    f_setcurpl2cpcache,
)
from ...memio import modcurpl as cp
from ...scdata import CUnit
from ...utils import EPD
from .unlimiterflag import IsUnlimiterOn


def EUDLoopList(  # noqa: N802
    header_offset, break_offset=None
) -> Iterator[tuple[c.EUDVariable, c.EUDVariable]]:
    blockname = "listloop"
    ut.EUDCreateBlock(blockname, header_offset)

    ptr, epd = f_dwepdread_epd(ut.EPD(header_offset))

    if break_offset is not None:
        cs.EUDWhileNot()(ptr == break_offset)
    else:
        cs.EUDWhile()([ptr > 0, ptr <= 0x7FFFFFFF])

    yield ptr, epd
    cs.EUDSetContinuePoint()
    epd += 1
    f_dwepdread_epd(epd, ret=[ptr, epd])
    cs.EUDEndWhile()

    ut.ep_assert(
        ut.EUDPopBlock(blockname)[1] is header_offset, _("listloop mismatch")
    )


def EUDLoopUnit() -> Iterator[tuple[c.EUDVariable, c.EUDVariable]]:  # noqa: N802
    ut.EUDCreateBlock("unitloop", 0x628430)

    ptr, epd = f_cunitepdread_epd(EPD(0x628430))

    if cs.EUDWhile()(ptr >= 1):
        yield ptr, epd
        cs.EUDSetContinuePoint()
        epd += 1
        f_cunitepdread_epd(epd, ret=[ptr, epd])
    cs.EUDEndWhile()

    ut.EUDPopBlock("unitloop")


class _UniqueIdentifier(c.EUDObject):
    def GetDataSize(self) -> int:  # noqa: N802
        return 4 + 336 * 1699

    def WritePayload(self, emitbuffer) -> None:  # noqa: N802
        emitbuffer.WriteDword(0)
        for _unit_data in range(1699):
            emitbuffer.WriteSpace(332)
            emitbuffer.WriteDword(0xFF)


def EUDLoopNewUnit(  # noqa: N802
    allowance: int = 2,
) -> Iterator[tuple[c.EUDVariable, c.EUDVariable]]:
    ut.ep_assert(isinstance(allowance, int))
    first_unit_ptr = EPD(0x628430)
    ut.EUDCreateBlock("newunitloop", "newlo")
    tos0 = c.EUDLightVariable(0)
    # tos0 << 0

    ptr, epd = f_cunitepdread_epd(first_unit_ptr)
    if cs.EUDWhileNot()(ptr == 0):
        c.VProc(
            epd,
            [
                c.SetMemory(0x6509B0, c.SetTo, 0xA5 // 4),
                epd.QueueAddTo(EPD(0x6509B0)),
            ],
        )
        uniq = f_bread_cp(0, 1)
        unique_identifier = _UniqueIdentifier()
        check_unique = c.DeathsX(cp.CP, c.Exactly, 0, 0, 0xFF)
        c.VProc(
            uniq,
            [
                c.SetMemory(
                    0x6509B0,
                    c.Add,
                    EPD(unique_identifier) - EPD(0x59CCA8) - 0xA5 // 4,
                ),
                uniq.SetDest(EPD(check_unique + 8)),
            ],
        )
        if cs.EUDIfNot()(check_unique):
            f_setcurpl2cpcache(uniq, uniq.SetDest(cp.CP))
            yield ptr, epd
        if cs.EUDElse()():
            tos0 += 1
            cs.EUDBreakIf(tos0.AtLeast(allowance))
        cs.EUDEndIf()
        cs.EUDSetContinuePoint()
        epd += 1
        f_cunitepdread_epd(epd, ret=[ptr, epd])
    cs.EUDEndWhile()

    f_setcurpl2cpcache([], tos0.SetNumber(0))
    ut.EUDPopBlock("newunitloop")


def EUDLoopUnit2() -> Iterator[tuple[c.EUDVariable, c.EUDVariable]]:  # noqa: N802
    """EUDLoopUnit보다 약간? 빠릅니다. 유닛 리스트를 따라가지 않고
    1700개 유닛을 도는 방식으로 작동합니다.
    """
    if IsUnlimiterOn():
        offset, is_dead = 0x4C // 4, c.MemoryXEPD(0, c.Exactly, 0, 0xFF00)
    else:
        offset, is_dead = 0x0C // 4, c.MemoryEPD(0, c.Exactly, 0)
    ptr = c.EUDVariable()
    epd = c.EUDVariable()
    continue_if = c.Forward()
    set_ptr, set_epd = ptr.SetNumber(0), epd.SetNumber(0)

    cs.DoActions(
        ptr.SetNumber(0x59CCA8),
        c.SetMemory(set_ptr + 20, c.SetTo, 0x59CCA8),
        epd.SetNumber(EPD(0x59CCA8)),
        c.SetMemory(is_dead + 4, c.SetTo, EPD(0x59CCA8) + offset),
        c.SetMemory(set_epd + 20, c.SetTo, EPD(0x59CCA8)),
    )
    if cs.EUDWhileNot()(ptr >= 0x59CCA8 + 336 * 1699 + 1):
        whileblock = ut.EUDPeekBlock("whileblock")[1]
        c.PushTriggerScope()
        continue_okay = c.Forward()
        continue_jump = c.RawTrigger(
            nextptr=whileblock["contpoint"],
            actions=c.SetNextPtr(continue_if, continue_okay),
        )
        c.PopTriggerScope()
        continue_if << c.RawTrigger(
            conditions=is_dead, actions=c.SetNextPtr(continue_if, continue_jump)
        )
        continue_okay << c.NextTrigger()
        yield ptr, epd

        cs.EUDSetContinuePoint()
        c.RawTrigger(
            nextptr=whileblock["loopstart"],
            actions=[
                c.SetMemory(is_dead + 4, c.Add, 84),
                c.SetMemory(set_ptr + 20, c.Add, 336),
                set_ptr,
                c.SetMemory(set_epd + 20, c.Add, 84),
                set_epd,
            ],
        )
    cs.EUDEndWhile()


def EUDLoopPlayerUnit(player) -> Iterator[tuple[c.EUDVariable, c.EUDVariable]]:  # noqa: N802
    player = c.EncodePlayer(player)
    first_player_unit = 0x6283F8
    ut.EUDCreateBlock("playerunitloop", first_player_unit)
    ptr, epd = f_cunitepdread_epd(ut.EPD(first_player_unit) + player)

    if cs.EUDWhile()(ptr >= 1):
        next_unit, nextptr = c.Forward(), c.Forward()
        # /*0x06C*/ BW::CUnit*  nextPlayerUnit;
        # f_cunitepdread_epd(
        #     epd + 0x6C // 4,
        #     ret=[EPD(next_unit) + 5, EPD(next_unit) + 13]
        # )
        c.VProc(
            epd,
            [
                c.SetMemory(0x6509B0, c.SetTo, 0x6C // 4),
                epd.QueueAddTo(EPD(0x6509B0)),
                f_cunitepdread_epd._frets[0].SetDest(EPD(next_unit) + 5),
                f_cunitepdread_epd._frets[1].SetDest(EPD(next_unit) + 13),
                c.SetMemory(f_cunitepdread_epd._nptr, c.SetTo, nextptr),
            ],
        )
        c.SetNextTrigger(f_cunitepdread_epd._fstart)
        nextptr << c.NextTrigger()
        yield ptr, epd
        cs.EUDSetContinuePoint()
        cs.DoActions(next_unit << ptr.SetNumber(0), epd.SetNumber(0))
    cs.EUDEndWhile()

    ut.EUDPopBlock("playerunitloop")


def EUDLoopBullet() -> Iterator[tuple[c.EUDVariable, c.EUDVariable]]:  # noqa: N802
    yield from EUDLoopList(0x64DEC4)


def EUDLoopSprite() -> Iterator[tuple[c.EUDVariable, c.EUDVariable]]:  # noqa: N802
    y_epd = c.EUDVariable()
    y_epd << ut.EPD(0x629688)

    ut.EUDCreateBlock("spriteloop", "sprlo")

    if cs.EUDWhile()(y_epd < ut.EPD(0x629688) + 256):
        ptr, epd = f_dwepdread_epd(y_epd)
        if cs.EUDWhile()(ptr >= 1):
            yield ptr, epd
            cs.EUDSetContinuePoint()
            epd += 1
            f_dwepdread_epd(epd, ret=[ptr, epd])
        cs.EUDEndWhile()
        y_epd += 1
    cs.EUDEndWhile()

    ut.EUDPopBlock("spriteloop")


def EUDLoopNewCUnit(allowance: int = 2) -> Iterator[CUnit]:  # noqa: N802
    for ptr, epd in EUDLoopNewUnit(allowance):
        yield CUnit(epd, ptr=ptr)


def EUDLoopCUnit() -> Iterator[CUnit]:  # noqa: N802
    """EUDLoopUnit보다 약간? 빠릅니다. 유닛 리스트를 따라가지 않고
    1700개 유닛을 도는 방식으로 작동합니다.
    """
    for ptr, epd in EUDLoopUnit2():
        yield CUnit(epd, ptr=ptr)


def EUDLoopPlayerCUnit(player) -> Iterator[CUnit]:  # noqa: N802
    for ptr, epd in EUDLoopPlayerUnit(player):
        yield CUnit(epd, ptr=ptr)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.


def T(x: int) -> int:  # noqa: N802
    x &= 0xFFFFFFFF
    for i in range(4):
        xsq = x * x
        x = (x * (xsq * (xsq * xsq + 1) + 1) + 0x8ADA4053) & 0xFFFFFFFF
    return x


def mix(x: int, y: int) -> int:
    return T(T(x) + y + 0x10F874F3) & 0xFFFFFFFF



from typing import Literal, TypeAlias

DefUpgradeDict: dict[str, int] = {
    "Terran Infantry Armor": 0,
    "Terran Vehicle Plating": 1,
    "Terran Ship Plating": 2,
    "Zerg Carapace": 3,
    "Zerg Flyer Carapace": 4,
    "Protoss Armor": 5,
    "Protoss Plating": 6,
    "Terran Infantry Weapons": 7,
    "Terran Vehicle Weapons": 8,
    "Terran Ship Weapons": 9,
    "Zerg Melee Attacks": 10,
    "Zerg Missile Attacks": 11,
    "Zerg Flyer Attacks": 12,
    "Protoss Ground Weapons": 13,
    "Protoss Air Weapons": 14,
    "Protoss Plasma Shields": 15,
    "U-238 Shells": 16,
    "Ion Thrusters": 17,
    "Burst Lasers (Unused)": 18,
    "Titan Reactor (SV +50)": 19,
    "Ocular Implants": 20,
    "Moebius Reactor (Ghost +50)": 21,
    "Apollo Reactor (Wraith +50)": 22,
    "Colossus Reactor (BC +50)": 23,
    "Ventral Sacs": 24,
    "Antennae": 25,
    "Pneumatized Carapace": 26,
    "Metabolic Boost": 27,
    "Adrenal Glands": 28,
    "Muscular Augments": 29,
    "Grooved Spines": 30,
    "Gamete Meiosis (Queen +50)": 31,
    "Metasynaptic Node (Defiler +50)": 32,
    "Singularity Charge": 33,
    "Leg Enhancements": 34,
    "Scarab Damage": 35,
    "Reaver Capacity": 36,
    "Gravitic Drive": 37,
    "Sensor Array": 38,
    "Gravitic Boosters": 39,
    "Khaydarin Amulet (HT +50)": 40,
    "Apial Sensors": 41,
    "Gravitic Thrusters": 42,
    "Carrier Capacity": 43,
    "Khaydarin Core (Arbiter +50)": 44,
    "Unused45": 45,
    "Unused46": 46,
    "Argus Jewel (Corsair +50)": 47,
    "Unused48": 48,
    "Argus Talisman (DA +50)": 49,
    "Unused50": 50,
    "Caduceus Reactor (Medic +50)": 51,
    "Chitinous Plating": 52,
    "Anabolic Synthesis": 53,
    "Charon Booster": 54,
    "Unused55": 55,
    "Unused56": 56,
    "Unused57": 57,
    "Unused58": 58,
    "Unused59": 59,
    "Unused60": 60,
}

DefaultUpgrade: TypeAlias = Literal[
    "Terran Infantry Armor",
    "Terran Vehicle Plating",
    "Terran Ship Plating",
    "Zerg Carapace",
    "Zerg Flyer Carapace",
    "Protoss Armor",
    "Protoss Plating",
    "Terran Infantry Weapons",
    "Terran Vehicle Weapons",
    "Terran Ship Weapons",
    "Zerg Melee Attacks",
    "Zerg Missile Attacks",
    "Zerg Flyer Attacks",
    "Protoss Ground Weapons",
    "Protoss Air Weapons",
    "Protoss Plasma Shields",
    "U-238 Shells",
    "Ion Thrusters",
    "Burst Lasers (Unused)",
    "Titan Reactor (SV +50)",
    "Ocular Implants",
    "Moebius Reactor (Ghost +50)",
    "Apollo Reactor (Wraith +50)",
    "Colossus Reactor (BC +50)",
    "Ventral Sacs",
    "Antennae",
    "Pneumatized Carapace",
    "Metabolic Boost",
    "Adrenal Glands",
    "Muscular Augments",
    "Grooved Spines",
    "Gamete Meiosis (Queen +50)",
    "Metasynaptic Node (Defiler +50)",
    "Singularity Charge",
    "Leg Enhancements",
    "Scarab Damage",
    "Reaver Capacity",
    "Gravitic Drive",
    "Sensor Array",
    "Gravitic Boosters",
    "Khaydarin Amulet (HT +50)",
    "Apial Sensors",
    "Gravitic Thrusters",
    "Carrier Capacity",
    "Khaydarin Core (Arbiter +50)",
    "Unused45",
    "Unused46",
    "Argus Jewel (Corsair +50)",
    "Unused48",
    "Argus Talisman (DA +50)",
    "Unused50",
    "Caduceus Reactor (Medic +50)",
    "Chitinous Plating",
    "Anabolic Synthesis",
    "Charon Booster",
    "Unused55",
    "Unused56",
    "Unused57",
    "Unused58",
    "Unused59",
    "Unused60",
]



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import itertools

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from ..memio import f_getcurpl, f_setcurpl, f_setcurpl2cpcache
from ..memio.rwcommon import br1, cw
from ..scdata import CurrentPlayer
from .cpprint import f_cpstr_print
from .strcommon import epd2s, ptr2s

color_codes = list(range(1, 32))
color_codes.remove(0x9)  # tab
color_codes.remove(0xA)  # linebreak
color_codes.remove(0xC)  # linebreak
color_codes.remove(0x12)  # right align
color_codes.remove(0x13)  # center align
color_v = c.EUDVariable(2)
color_code = b"\x02"


@c.EUDFunc
def _cpchar_addstr():
    if cs.EUDInfLoop()():
        b1 = br1.readbyte()
        cs.EUDBreakIf(b1 == 0)
        cw.writebyte(color_v)
        cw.writebyte(b1)
        if cs.EUDIf()(b1 <= 0x7F):
            cw.flushdword()
        if cs.EUDElse()():
            b2 = br1.readbyte()
            cw.writebyte(b2)
            if cs.EUDIf()(b1 <= 0xDF):  # Encode as 2-byte
                cw.flushdword()
            if cs.EUDElse()():  # 3-byte
                cw.writebyte(br1.readbyte())
            cs.EUDEndIf()
        cs.EUDEndIf()
    cs.EUDEndInfLoop()


def f_cpchar_addstr(src):
    br1.seekoffset(src)
    _cpchar_addstr()


def f_cpchar_addstr_epd(src):
    br1.seekepd(src)
    _cpchar_addstr()


@c.EUDFunc
def f_cpchar_adddw(number):
    skipper = [c.Forward() for _ in range(9)]
    ch = [0] * 10

    for i in range(10):  # Get digits
        number, ch[i] = c.f_div(number, 10)
        if i != 9:
            cs.EUDJumpIf(number == 0, skipper[i])

    for i in range(9, -1, -1):  # print digits
        if i != 9:
            skipper[i] << c.NextTrigger()
        cs.DoActions(
            c.SetDeaths(
                CurrentPlayer,
                c.SetTo,
                color_v + ch[i] * 256 + (0x0D0D3000),
                0,
            ),
            c.AddCurrentPlayer(1),
        )


def f_cpchar_print(*args, EOS=True, encoding="UTF-8"):  # noqa: N803
    global color_code
    args = ut.FlattenList(args)

    if encoding.upper() == "UTF-8":
        encode_func = ut.u2utf8
    else:
        encode_func = ut.u2b

    for arg in args:
        if isinstance(arg, bytes):
            arg = arg.decode(encoding)
        if isinstance(arg, str):
            bytestring = b""
            new_color_code = color_code
            for char in arg:
                char = encode_func(char)
                if ut.b2i1(char) in color_codes:
                    new_color_code = char
                    continue
                while len(char) < 3:
                    char = char + b"\r"
                bytestring = bytestring + new_color_code + char
            if not bytestring:
                bytestring = color_code + b"\r\r\r"
            f_cpstr_print(bytestring, EOS=False)
            if color_code != new_color_code:
                color_code = new_color_code
                cs.DoActions(color_v.SetNumber(ut.b2i1(color_code)))
        elif ut.isUnproxyInstance(arg, ptr2s):
            f_cpchar_addstr(arg._value)
        elif ut.isUnproxyInstance(arg, epd2s):
            f_cpchar_addstr_epd(arg._value)
        elif c.IsEUDVariable(arg) or c.IsConstExpr(arg):
            f_cpchar_adddw(arg)
        else:
            f_cpstr_print(arg, EOS=False, encoding=encoding)
    if EOS:
        cs.DoActions(c.SetDeaths(CurrentPlayer, c.SetTo, 0, 0))


_TextFX_dict = dict()
id_codes = [
    "\x18",
    "\x02",
    "\x03",
    "\x04",
    "\x06",
    "\x07",
    "\x08",
    "\x0D",
    "\x0E",
    "\x0F",
    "\x10",
    "\x11",
    "\x15",
    "\x16",
    "\x17",
    "\x19",
    "\x1A",
    "\x1B",
    "\x1C",
    "\x1D",
    "\x1E",
    "\x1F",
]
id_gen = itertools.cycle(itertools.product(id_codes, repeat=6))


def _add_textfx_timer(tag):
    if tag not in _TextFX_dict:
        _TextFX_dict[tag] = (
            c.EUDVariable(),
            c.EUDLightVariable(),
            ut.u2b("".join(next(id_gen))),
        )


def get_textfx_timer(tag):
    return _TextFX_dict[tag]


def TextFX_SetTimer(tag, modtype, value):  # noqa: N802
    _add_textfx_timer(tag)
    timer, _, _ = get_textfx_timer(tag)
    cs.DoActions(c.SetMemory(timer.getValueAddr(), modtype, value))


@c.EUDFunc
def _remove_textfx(o0, o1, e0, e1):
    textptr = c.EUDVariable()
    e0t, e1t, e_fail, o0t, o1t, o_fail = (c.Forward() for _ in range(6))
    setptr_o, setptr_e = c.Forward(), c.Forward()

    c.VProc(
        [o0, o1, e0, e1],
        [
            textptr.SetNumber(-1),
            c.SetMemory(0x6509B0, c.SetTo, ut.EPD(0x640B60 + 218 * 11)),
            c.SetMemory(setptr_o + 20, c.SetTo, 10),
            c.SetMemory(setptr_e + 20, c.SetTo, 11),
            o0.SetDest(ut.EPD(o0t + 16)),
            o1.SetDest(ut.EPD(o1t + 16)),
            e0.SetDest(ut.EPD(e0t + 16)),
            e1.SetDest(ut.EPD(e1t + 16)),
        ],
    )
    if cs.EUDLoopN()(6):
        incr_e, incr_o = c.Forward(), c.Forward()
        e0t << c.RawTrigger(
            nextptr=incr_e,
            conditions=c.DeathsX(CurrentPlayer, c.Exactly, 0, 0, 0xFFFF0000),
            actions=[c.SetMemory(0x6509B0, c.Add, 1), c.SetNextPtr(e0t, e1t)],
        )
        e1t << c.RawTrigger(
            conditions=c.Deaths(CurrentPlayer, c.Exactly, 0, 0),
            actions=[
                c.SetDeaths(CurrentPlayer, c.SetTo, 0, 0),
                c.SetMemory(0x6509B0, c.Subtract, 1),
                c.SetDeathsX(CurrentPlayer, c.SetTo, 0, 0, 0xFFFF0000),
                setptr_e << textptr.SetNumber(0),
                c.SetNextPtr(e1t, incr_e),
            ],
        )
        e_fail << c.RawTrigger(actions=c.SetMemory(0x6509B0, c.Subtract, 1))
        incr_e << c.RawTrigger(
            actions=[
                c.SetNextPtr(e0t, incr_e),
                c.SetNextPtr(e1t, e_fail),
                c.SetMemory(0x6509B0, c.Subtract, 216 // 4),
                c.SetMemory(setptr_e + 20, c.Subtract, 2),
            ]
        )
        o0t << c.RawTrigger(
            nextptr=incr_o,
            conditions=c.Deaths(CurrentPlayer, c.Exactly, 0, 0),
            actions=[c.SetMemory(0x6509B0, c.Add, 1), c.SetNextPtr(o0t, o1t)],
        )
        o1t << c.RawTrigger(
            conditions=c.DeathsX(CurrentPlayer, c.Exactly, 0, 0, 0xFFFF),
            actions=[
                c.SetDeaths(CurrentPlayer, c.SetTo, 0, 0),
                c.SetMemory(0x6509B0, c.Subtract, 1),
                c.SetDeaths(CurrentPlayer, c.SetTo, 0, 0),
                setptr_o << textptr.SetNumber(0),
                c.SetNextPtr(o1t, incr_o),
            ],
        )
        o_fail << c.RawTrigger(actions=c.SetMemory(0x6509B0, c.Subtract, 1))
        incr_o << c.RawTrigger(
            actions=[
                c.SetNextPtr(o0t, incr_o),
                c.SetNextPtr(o1t, o_fail),
                c.SetMemory(0x6509B0, c.Subtract, 220 // 4),
                c.SetMemory(setptr_o + 20, c.Subtract, 2),
            ]
        )
    cs.EUDEndLoopN()
    f_setcurpl2cpcache()
    return textptr


def TextFX_Remove(tag):  # noqa: N802
    ut.ep_assert(tag is not None, "tag should not be None")
    _add_textfx_timer(tag)
    _, _, identifier = get_textfx_timer(tag)
    o0 = ut.b2i4(identifier[0:4])
    o1 = ut.b2i2(identifier[4:6])
    e0 = ut.b2i2(identifier[0:2]) * 0x10000
    e1 = ut.b2i4(identifier[2:6])
    return _remove_textfx(o0, o1, e0, e1)


_check_cp = c.Forward()
_is_below_start = c.Forward()
_cpbelowbuffer = c.EUDLightBool()


def _is_cp_less_than_start(actions):
    global _is_below_start
    if not _is_below_start.IsSet():
        c.PushTriggerScope()
        _is_below_start << c.RawTrigger(
            conditions=[_check_cp << c.Memory(0x6509B0, c.AtMost, 0)],
            actions=_cpbelowbuffer.Set(),
        )
        c.PopTriggerScope()
    _next = c.Forward()
    actions = ut.FlattenList(actions)
    c.RawTrigger(
        nextptr=_is_below_start,
        actions=[
            *actions,
            _cpbelowbuffer.Clear(),
            c.SetNextPtr(_is_below_start, _next),
        ],
    )
    _next << c.NextTrigger()


def _r2l(colors, colors_dict={}):
    try:
        _f = colors_dict[colors]
    except KeyError:

        @c.EUDFunc
        def _f():
            _jump, _isend, _end = (c.Forward() for _ in range(3))
            ret = c.EUDVariable()
            _is_cp_less_than_start([ret.SetNumber(1), c.SetNextPtr(_isend, _jump)])
            _isend << c.RawTrigger(
                conditions=_cpbelowbuffer.IsSet(),
                actions=[ret.SetNumber(0), c.SetNextPtr(_isend, _end)],
            )
            _jump << c.NextTrigger()
            for color in reversed(colors):
                _is_cp_less_than_start([])
                c.RawTrigger(
                    conditions=_cpbelowbuffer.IsCleared(),
                    actions=[
                        c.SetDeathsX(CurrentPlayer, c.SetTo, color, 0, 0xFF),
                        c.AddCurrentPlayer(-1),
                    ],
                )
            _end << c.NextTrigger()
            return ret

        colors_dict[colors] = _f
    return _f()


def _textfx_print(*args, identifier, encoding="UTF-8"):
    f_cpstr_print(identifier, EOS=False, encoding=encoding)
    global color_code
    color_code = b"\x02"
    color_v << 2

    args = ut.FlattenList(args)
    for arg in args:
        if isinstance(arg, str) and "\n" in arg:
            line = arg.split("\n")
            for s in line[:-1]:
                f_cpchar_print(s, EOS=False, encoding=encoding)
                f_cpstr_print(
                    identifier + b"\n" + identifier,
                    EOS=False,
                    encoding=encoding,
                )
            f_cpchar_print(line[-1], EOS=False, encoding=encoding)
        else:
            f_cpchar_print(arg, EOS=False, encoding=encoding)

    cs.DoActions(c.SetDeaths(CurrentPlayer, c.SetTo, 0, 0))


def TextFX_FadeIn(  # noqa: N802
    *args, color=None, wait=1, reset=True, tag=None, encoding="UTF-8"
):
    """Print multiple string / number and apply color from Left To Right

    Keyword arguments:
    color -- tuple of color codes (default 0x03, 0x04, 0x05, 0x14)
    wait  -- time interval between effect (default 1)
    reset -- automatically reset when didn't run for a moment (default True)
    tag   -- internal tag and identifier (default: vargs)
    """
    from ..eudlib.utilf.gametick import f_getgametick

    if color is None:
        color = (0x03, 0x04, 0x05, 0x14)
    if tag is None:
        if len(args) == 1:
            tag = args[0]
        else:
            tag = args

    _add_textfx_timer(tag)
    timer, counter, identifier = get_textfx_timer(tag)

    start = f_getcurpl()
    c.VProc(
        start,
        [
            start.AddNumber(1),
            start.SetDest(ut.EPD(_check_cp) + 2),
        ],
    )
    _textfx_print(*args, identifier=identifier, encoding=encoding)
    f_setcurpl(start, actions=[start.AddNumber(2 - len(color))], set_modifier=False)

    if reset is True:
        check_gametick = c.Forward()
        if cs.EUDIf()([check_gametick << c.Memory(0x57F23C, c.AtLeast, 0)]):
            gametick = f_getgametick()
            c.VProc(
                gametick,
                [
                    timer.SetNumber(0),
                    gametick.AddNumber(1),
                    gametick.SetDest(ut.EPD(check_gametick) + 2),
                ],
            )
        cs.EUDEndIf()

    _is_finished, _draw_color, _end = (c.Forward() for _ in range(3))
    ret = c.EUDVariable()

    cs.DoActions(
        c.SetMemory(check_gametick + 8, c.Add, 1) if reset is True else [],
        c.SetNextPtr(_is_finished, _draw_color),
        c.AddCurrentPlayer(timer),
    )
    _is_finished << c.RawTrigger(
        conditions=[
            c.Deaths(CurrentPlayer, c.Exactly, 0, 0),
            timer.AtLeast(2 + len(color)),
        ],
        actions=[
            ret.SetNumber(0),
            counter.SetNumber(0),
            c.SetNextPtr(_is_finished, _end),
        ],
    )
    _draw_color << c.RawTrigger(
        actions=[
            ret.SetNumber(1),
            counter.AddNumber(1),
            c.AddCurrentPlayer(len(color) - 1),
        ]
    )
    _r2l(color)
    c.RawTrigger(
        conditions=counter.AtLeast(max(wait, 1)),
        actions=[counter.SetNumber(0), timer.AddNumber(1)],
    )
    _end << c.NextTrigger()
    return ret


def TextFX_FadeOut(  # noqa: N802
    *args, color=None, wait=1, reset=True, tag=None, encoding="UTF-8"
):
    """Print multiple string / number and apply color from Left To Right

    Keyword arguments:
    color -- tuple of color codes (default 0x03, 0x04, 0x05, 0x14)
    wait  -- time interval between effect (default 1)
    reset -- automatically reset when didn't run for a moment (default True)
    tag   -- internal tag and identifier (default: vargs)
    """
    from ..eudlib.utilf.gametick import f_getgametick

    if color is None:
        color = (0x03, 0x04, 0x05, 0x14)
    if tag is None:
        if len(args) == 1:
            tag = args[0]
        else:
            tag = args

    _add_textfx_timer(tag)
    timer, counter, identifier = get_textfx_timer(tag)

    start = f_getcurpl()
    c.VProc(
        start,
        [
            start.AddNumber(1),
            start.SetDest(ut.EPD(_check_cp) + 2),
        ],
    )
    _textfx_print(*args, identifier=identifier, encoding=encoding)

    if reset is True:
        check_gametick = c.Forward()
        if cs.EUDIf()([check_gametick << c.Memory(0x57F23C, c.AtLeast, 0)]):
            gametick = f_getgametick()
            c.VProc(
                gametick,
                [
                    timer.SetNumber(0),
                    gametick.AddNumber(1),
                    gametick.SetDest(ut.EPD(check_gametick) + 2),
                ],
            )
        cs.EUDEndIf()

    does_last_color_overwrite = color[-1] in {0, 5, 0x14}
    increment_and_jump, paint_last_color = c.Forward(), c.Forward()

    cs.DoActions(
        c.SetMemory(check_gametick + 8, c.Add, 1) if reset is True else [],
        c.AddCurrentPlayer((len(color) - 1) - timer),
        c.SetNextPtr(increment_and_jump, paint_last_color)
        if does_last_color_overwrite
        else [],
    )
    ret = _r2l(color)
    if does_last_color_overwrite:
        skip = c.Forward()
        increment_and_jump << c.RawTrigger(
            conditions=ret.Exactly(1),
            actions=[
                counter.AddNumber(1),
                c.SetNextPtr(increment_and_jump, skip),
            ],
        )
        paint_last_color << c.VProc(
            start,
            [
                start.AddNumber(1),
                start.SetDest(ut.EPD(0x6509B0)),
            ],
        )
        f_setcurpl2cpcache(
            [], c.SetDeathsX(CurrentPlayer, c.SetTo, color[-1], 0, 0xFF)
        )
        skip << c.NextTrigger()
    else:
        c.RawTrigger(conditions=ret.Exactly(1), actions=counter.AddNumber(1))
    c.RawTrigger(
        conditions=counter.AtLeast(max(wait, 1)),
        actions=[counter.SetNumber(0), timer.AddNumber(1)],
    )
    return ret



# Copyright 2024 by zzt (Defender).
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

# ruff: noqa: N815
from typing import ClassVar

from .. import utils as ut
from ..core.rawtrigger.consttype import ConstType
from ..core.rawtrigger.strenc import EncodeSprite
from ..core.rawtrigger.typehint import _Sprite
from ..localize import _
from .offsetmap import BoolMember, EPDOffsetMap, ImageMember


class Sprite(EPDOffsetMap, ConstType):
    __slots__ = ()
    # Read only data skipped
    image: ClassVar[ImageMember] = ImageMember("array", 0x666160)
    # hpBarSize: ClassVar[ArrayMember] = ArrayMember(0x665E50, Mk.BYTE)
    # hpBarSize starts on Sprites.dat ID 130
    # unknownFlag: ClassVar[ArrayMember] = ArrayMember(0x666570, Mk.BYTE)
    isVisible: ClassVar[BoolMember] = BoolMember("array", 0x665C48)
    # selectionCircle: ClassVar[ArrayMember] = ArrayMember(0x665AC0, Mk.BYTE)
    # selectionCircle and selectionVerticalOffset start on Sprites.dat ID 130
    # selectionVerticalOffset: ClassVar[ArrayMember] = ArrayMember(0x665FD8, Mk.BYTE)

    @ut.classproperty
    def range(self):
        return (0, 516, 1)

    @classmethod
    def cast(cls, _from: _Sprite):
        if isinstance(_from, ConstType) and not isinstance(_from, cls):
            raise ut.EPError(_('"{}" is not a {}').format(_from, cls.__name__))
        return super().cast(_from)

    def __init__(self, initval: _Sprite) -> None:
        super().__init__(EncodeSprite(initval))



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ..core.eudstruct.vararray import _lv
from .byterw import EUDByteReader, EUDByteStream, EUDByteWriter
from .cpbyterw import CPByteWriter

lv = _lv
cw = CPByteWriter()
br1 = EUDByteReader()
br2 = EUDByteReader()
# br3 = bm.EUDByteReader()
bw1 = EUDByteWriter()
# bw2 = bm.EUDByteWriter()
# bw3 = bm.EUDByteWriter()
bs1 = EUDByteStream()



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import gettext
import locale
import os
import sys
from types import TracebackType
from typing import Any

# Determine locale from environment or default
lc = None
if lc is None:
    lc, _e = locale.getdefaultlocale()

lang = None if lc is None else (lc,)

# Set up locale path and translation
locale_path = os.path.dirname(__file__)
t = gettext.translation("eudplib", locale_path, languages=lang, fallback=True)
if not isinstance(t, gettext.GNUTranslations):
    # Adjust locale_path if necessary
    locale_path = os.path.dirname(os.path.dirname(os.path.dirname(locale_path)))
    t = gettext.translation("eudplib", locale_path, languages=lang, fallback=True)

# Set up gettext function
_ = t.gettext

# Preserve original hooks
original_excepthook = sys.excepthook
original_unraisablehook = sys.unraisablehook


def excepthook(
    type_: type[BaseException],
    value: BaseException,
    traceback: TracebackType | None,
) -> None:
    # Localize exception messages
    v = list(value.args)
    for i, s in enumerate(v):
        if isinstance(s, str):
            v[i] = t.gettext(s)
    value.args = tuple(v)
    original_excepthook(type_, value, traceback)


def unraisablehook(unraisable: Any) -> None:
    # Localize unraisable error messages
    err_msg = unraisable.err_msg or "Exception ignored in"
    err_msg = t.gettext(err_msg)
    obj = unraisable.object
    print(f"{err_msg}: {obj!r}")


# Set custom hooks
sys.excepthook = excepthook
sys.unraisablehook = unraisablehook

__all__ = ["_"]



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

# Just run _eudvsupport. This won't be visible in user side.
from . import _eudvsupport  # FIXME: remove hacky way



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Callable

from ..variable import EUDVariable
from .bitwise import f_bitlshift, f_bitrshift
from .muldiv import _quot, _rem, f_mul


def _def_cls_method(name: str, f: Callable) -> None:
    f.__name__ = f"EUDVariable.{name}"
    setattr(EUDVariable, name, f)


def _def_bin_operator(name: str, f: Callable) -> None:
    _def_cls_method(name, f)
    _def_cls_method(f"__r{name[2:]}", lambda self, lhs: f(lhs, self))


def _def_inplace_operator(name: str, f: Callable) -> None:
    def iop(self, other):
        rvalue = self._rvalue
        ret = f(self, other)
        if not rvalue:
            ret.makeL()
        return ret

    _def_cls_method(name, iop)


_def_bin_operator("__mul__", lambda x, y: f_mul(x, y))
_def_bin_operator("__floordiv__", lambda x, y: _quot(x, y))
_def_bin_operator("__mod__", lambda x, y: _rem(x, y))
_def_bin_operator("__rshift__", lambda x, y: f_bitrshift(x, y))
_def_inplace_operator("__imul__", lambda x, y: f_mul(x, y, ret=[x]))
_def_inplace_operator("__ifloordiv__", lambda x, y: _quot(x, y, ret=[x]))
_def_inplace_operator("__imod__", lambda x, y: _rem(x, y, ret=[x]))
_def_inplace_operator("__ilshift__", lambda a, b: f_bitlshift(a, b, ret=[a]))
_def_inplace_operator("__irshift__", lambda a, b: f_bitrshift(a, b, ret=[a]))
_def_cls_method("__rlshift__", lambda a, b: f_bitlshift(b, a))

# Shift operator is reserved for assigning, so we won't overload them.



import types

import eudplib

# mode = 1  # Get functions
# mode = 2  # Get globals
mode = 3  # Everything

module_to_doc = eudplib

exclude_types = [dict, str, types.ModuleType]
eudftypes = [eudplib.core.eudfunc.eudfuncn.EUDFuncN]
if mode == 1:
    allowed_type = [
        *eudftypes,
        type,
        types.FunctionType,
        eudplib.core.rawtrigger.constenc._Kills,
    ]
elif mode == 2:
    allowed_type = [
        int,
        eudplib.core.rawtrigger.consttype.ConstType,
        eudplib.core.rawtrigger.constenc._Kills,
    ]
else:
    allowed_type = [
        *eudftypes,
        int,
        eudplib.core.rawtrigger.consttype.ConstType,
        type,
        types.FunctionType,
        eudplib.core.rawtrigger.constenc._Kills,
    ]
exclude_names = ["__loader__", "__path__", "__spec__"]

# Document module
nlist = []
for name, value in module_to_doc.__dict__.items():
    # Exclude list
    if name in exclude_names or type(value) in exclude_types:
        continue

    # Undocumented -> ignore
    if not any(isinstance(value, t) for t in allowed_type):
        continue

    else:
        documented = True

    nlist.append(name)

if mode != 2:
    nlist.extend(
        ["beforeTriggerExec", "afterTriggerExec", "onPluginStart", "settings"]
    )
print("\n==================================\n")
nlist.sort()
nlist = [f'"{name}", ' for name in nlist]
col = 4
s: list[str] = []


def flush():
    global col
    print("    " + "".join(s))
    col = 4
    s.clear()


for n in nlist:
    if col + len(n) >= 80:
        flush()
    s.append(n)
    col += len(n)
flush()



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import struct

from typing_extensions import Self

from .. import core as c


class EUDGrp(c.EUDObject):
    """Object for GRP

    Starcraft modifies GRP in certain way before it is used ingame. This object
    emulates that modification so that SC recognizes GRP correctly.
    """

    def __new__(cls, *args, **kwargs) -> Self:
        return super().__new__(cls)

    def __init__(self, content) -> None:
        super().__init__()
        if isinstance(content, str):
            with open(content, "rb") as file:
                content = file.read()
        self._content = content

    def Evaluate(self):  # noqa: N802
        return c.GetObjectAddr(self) + 2

    def GetDataSize(self):  # noqa: N802
        return len(self._content) + 2

    def WritePayload(self, buf):  # noqa: N802
        buf.WriteBytes(b"\0\0")  # 2byte padding to align dwords at (*)

        # fill in grp header
        b = self._content
        fn, w, h = struct.unpack("<HHH", b[0:6])
        buf.WriteWord(fn)
        buf.WriteWord(w)
        buf.WriteWord(h)

        # fill in grp frame headers table
        selfaddr = self.Evaluate()

        for i in range(fn):
            fhoffset = 6 + 8 * i
            xoff, yoff, w, h, lto = struct.unpack(
                "<BBBBI", b[fhoffset : fhoffset + 8]
            )
            buf.WriteByte(xoff)
            buf.WriteByte(yoff)
            buf.WriteByte(w)
            buf.WriteByte(h)
            buf.WriteDword(lto + selfaddr)  # (*)

        buf.WriteBytes(b[6 + 8 * fn :])



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .binio import b2i1, b2i2, b2i4, bits, i2b1, i2b2, i2b4
from .blockstru import (
    EUDCreateBlock,
    EUDGetBlockList,
    EUDGetLastBlock,
    EUDGetLastBlockOfName,
    EUDPeekBlock,
    EUDPopBlock,
)
from .eperror import (
    EPError,
    EPWarning,
    TriggerScopeError,
    ep_assert,
    ep_eprint,
    ep_warn,
)
from .etc import (
    EPD,
    Assignable2List,
    FlattenIter,
    FlattenList,
    List2Assignable,
    SCMD2Text,
    _rand_lst,
    cachedfunc,
    classproperty,
    find_data_file,
)
from .exprproxy import ExprProxy, isUnproxyInstance, unProxy
from .ubconv import b2u, b2utf8, u2b, u2utf8

__all__ = [
    "b2i1",
    "b2i2",
    "b2i4",
    "bits",
    "i2b1",
    "i2b2",
    "i2b4",
    "EUDCreateBlock",
    "EUDGetBlockList",
    "EUDGetLastBlock",
    "EUDGetLastBlockOfName",
    "EUDPeekBlock",
    "EUDPopBlock",
    "EPError",
    "EPWarning",
    "TriggerScopeError",
    "ep_assert",
    "ep_eprint",
    "ep_warn",
    "EPD",
    "Assignable2List",
    "FlattenIter",
    "FlattenList",
    "List2Assignable",
    "SCMD2Text",
    "_rand_lst",
    "cachedfunc",
    "classproperty",
    "find_data_file",
    "ExprProxy",
    "isUnproxyInstance",
    "unProxy",
    "b2u",
    "b2utf8",
    "u2b",
    "u2utf8",
]
